--- old/usr/src/uts/common/fs/nfs/nfs_srv.c	2012-08-04 02:30:38.070492774 +0400
+++ new/usr/src/uts/common/fs/nfs/nfs_srv.c	2012-08-04 02:30:37.458492758 +0400
@@ -326,6 +326,51 @@
 	return (&args->saa_fh);
 }
 
+/* Change and release @exip and @vpp only in success */
+int
+rfs_cross_mnt(vnode_t **vpp, struct exportinfo **exip)
+{
+	struct exportinfo *exi;
+	vnode_t *vp;
+	fid_t fid;
+	int error;
+
+	vp = *vpp;
+
+	/* traverse() releases argument in success */
+	VN_HOLD(*vpp);
+
+	if ((error = traverse(&vp)) != 0) {
+		VN_RELE(*vpp);
+		return (error);
+	}
+
+	bzero(&fid, sizeof (fid));
+	fid.fid_len = MAXFIDSZ;
+	error = VOP_FID(vp, &fid, NULL);
+	if (error) {
+		VN_RELE(vp);
+		return (error);
+	}
+
+	exi = checkexport(&vp->v_vfsp->vfs_fsid, &fid);
+	if (exi == NULL ||
+	    (exi->exi_export.ex_flags & EX_NOHIDE) == 0) {
+		/* It is not error, just subdir is not exported
+		 * or "nohide" is not set
+		 */
+		VN_RELE(vp);
+	} else {
+		/* go to submount */
+		exi_rele(*exip);
+		*exip = exi;
+
+		VN_RELE(*vpp);
+		*vpp = vp;
+	}
+	return (0);
+}
+
 /*
  * Directory lookup.
  * Returns an fhandle and file attributes for file name in a directory.
@@ -399,6 +444,8 @@
 		return;
 	}
 
+	exi_hold(exi);
+
 	/*
 	 * If the public filehandle is used then allow
 	 * a multi-component lookup, i.e. evaluate
@@ -409,9 +456,16 @@
 	 * which is OK as long as the filesystem is exported.
 	 */
 	if (PUBLIC_FH2(fhp)) {
+		struct exportinfo *new;
+
 		publicfh_flag = TRUE;
-		error = rfs_publicfh_mclookup(name, dvp, cr, &vp, &exi,
+		error = rfs_publicfh_mclookup(name, dvp, cr, &vp, &new,
 		    &sec);
+
+		if (error == 0) {
+			exi_rele(exi);
+			exi = new;
+		}
 	} else {
 		/*
 		 * Do a normal single component lookup.
@@ -423,6 +477,11 @@
 	if (name != da->da_name)
 		kmem_free(name, MAXPATHLEN);
 
+	if (error == 0 && vn_ismntpt(vp)) {
+		error = rfs_cross_mnt(&vp, &exi);
+		if (error)
+			VN_RELE(vp);
+	}
 
 	if (!error) {
 		va.va_mask = AT_ALL;	/* we want everything */
@@ -451,14 +510,10 @@
 
 	VN_RELE(dvp);
 
-	/*
-	 * If publicfh_flag is true then we have called rfs_publicfh_mclookup
-	 * and have obtained a new exportinfo in exi which needs to be
-	 * released. Note the the original exportinfo pointed to by exi
-	 * will be released by the caller, comon_dispatch.
+	/* The passed argument exportinfo is released by the
+	 * caller, comon_dispatch
 	 */
-	if (publicfh_flag && exi != NULL)
-		exi_rele(exi);
+	exi_rele(exi);
 
 	/*
 	 * If it's public fh, no 0x81, and client's flavor is
